//  DNR2 Ver.0.7 : Dynamic Noise Reduction 2 for Avisynth
//  Copyright (C) 2002 kiraru2002
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA,
//
//  or visit http://www.gnu.org/copyleft/gpl.html .
//
//------------------------------------------------------------------------------------
// kiraru2002's web site  http://members.tripod.co.jp/kiraru2002/
//
// This plug-in is ported from the VirtualDub's plug-ins "DNR1" and
// "DNR2" which Steven Don and Avery Lee created.
// You MUST NOT contact to Steven Don and Avery Lee for any trouble
// which happened using this plug-in, and for any things about this plug-in.
//------------------------------------------------------------------------------------
#include "Dnr.h"

DNR::DNR(PClip _child,	int _threshY_hi, int _threshY_lo,
						int _threshU_hi, int _threshU_lo,
						int _threshV_hi, int _threshV_lo,
						int _dnr_type,
						IScriptEnvironment* env)
		:	GenericVideoFilter(_child),
			threshY_hi(_threshY_hi),threshY_lo(_threshY_lo),
			threshU_hi(_threshU_hi),threshU_lo(_threshU_lo),
			threshV_hi(_threshV_hi),threshV_lo(_threshV_lo),dnr_type(_dnr_type)
{
	if (vi.IsRGB32())
	{
		if (threshU_hi == -1)	threshU_hi = threshY_hi;
		if (threshV_hi == -1)	threshV_hi = threshY_hi;

		if (threshY_lo == -1)	threshY_lo = (threshY_hi>>1) + 1;	//R
		if (threshU_lo == -1)	threshU_lo = (threshU_hi>>1) + 1;	//G
		if (threshV_lo == -1)	threshV_lo = (threshV_hi>>1) + 1;	//B

		thresh1 = 0x0000000100000000i64 * threshY_lo
					+ 0x0000000000010000i64 * threshU_lo
					+ 0x0000000000000001i64 * threshV_lo;

		thresh2 = 0x0000000100000000i64 * threshY_hi
					+ 0x0000000000010000i64 * threshU_hi
					+ 0x0000000000000001i64 * threshV_hi;
	} else {
		if (threshU_hi == -1)	threshU_hi = threshY_hi * 3 / 4;
		if (threshV_hi == -1)	threshV_hi = threshU_hi;

		if (threshY_lo == -1)	threshY_lo = (threshY_hi>>1) + 1;
		if (threshU_lo == -1)	threshU_lo = (threshU_hi>>1) + 1;
		if (threshV_lo == -1)	threshV_lo = (threshV_hi>>1) + 1;

		if (vi.IsYUY2())
		{
			thresh1 = 0x0000000100000001i64 * threshY_lo
						+ 0x0000000000010000i64 * threshU_lo
						+ 0x0001000000000000i64 * threshV_lo;

			thresh2 = 0x0000000100000001i64 * threshY_hi
						+ 0x0000000000010000i64 * threshU_hi
						+ 0x0001000000000000i64 * threshV_hi;
		}
	}

	OldFrameNo = -9999;

	divTable = NULL;
	if( !(env->GetCPUFlags() & CPUF_MMX) ) {
		if (!(divTable = new unsigned char [766])) return;
		for (unsigned int i = 0; i < 766; i++) 		divTable[i] = i / 3;
	}
		
	fc = child->GetFrame(0, env);
	if (!(old_data = new unsigned char [vi.RowSize() * fc->GetHeight()])) return;
}
					

DNR::~DNR()
{
	if(old_data)	{ delete [] old_data; old_data=NULL; }
	if(divTable)	{ delete [] divTable; divTable = NULL; }
}


//MMX optimised version
void __declspec(naked) dodnr2MMX(unsigned char *dst, unsigned char *src,	int w, int h,
								 int dstmodulo,  __int64 thresh1, __int64 thresh2)
{
	static const __int64 bythree = 0x5555555555555555i64;
	static const __int64 round2 =  0x0002000200020002i64;
	static const __int64 three =   0x0003000300030003i64;

	__asm {
		push	ebp
		push	edi
		push	esi
		push	ebx

		mov		edi,[esp+4+16]		;dst
		mov		esi,[esp+8+16]		;src
		mov		edx,[esp+12+16]		;w
		mov		ecx,[esp+16+16]		;h
		xor		eax,eax
		mov		ebx,[esp+20+16]		;dstmodulo
		pxor	mm7,mm7
		movq	mm5,[esp+24+16]		;thresh1
		movq	mm6,[esp+32+16]		;thresh2

yloop:
		mov		ebp,edx
xloop:
		movd	mm0,[esi]		;previous
		pxor	mm7,mm7

		movd	mm1,[edi]		;current
		punpcklbw	mm0,mm7

		punpcklbw	mm1,mm7
		movq	mm2,mm0

		movq	mm4,mm1			;mm4 = current
		movq	mm3,mm1

		movq	mm7,mm0
		paddw	mm4,mm4			;mm4 = current * 2

		pmullw	mm0,three		;mm0 = previous * 3
		psubusb	mm2,mm1

		paddw	mm4,mm7			;mm4 = current * 2 + previous
		psubusb	mm3,mm7

		pmulhw	mm4,bythree		;mm4 = (current * 2 + previous) / 3
		por		mm2,mm3

		movq	mm3,mm2
		paddw	mm0,mm1			;mm0 = previous * 3 + current

		paddw	mm0,round2		;mm0 = previous * 3 + current + 2
		pcmpgtw	mm2,mm5			;set if diff > thresh1

		pcmpgtw	mm3,mm6			;set if diff > thresh2
		psrlw	mm0,2			;mm0 = (previous * 3 + current + 2) / 4


		;	mm2		mm3		meaning						mm1		mm0		mm4
		;	FALSE	FALSE	diff <= thresh1				off		on		off
		;	FALSE	TRUE	impossible
		;	TRUE	FALSE	thresh1 < diff <= thresh2	off		off		on
		;	TRUE	TRUE	diff > thresh2				on		off		off

		pand	mm1,mm3			;keep pixels exceeding threshold2
		pand	mm4,mm2			;	average pixels <= threshold2...
		pandn	mm2,mm0			;replace pixels below threshold1
		pandn	mm3,mm4			;	but >= threshold1...
		por		mm1,mm2
		add		esi,4
		por		mm1,mm3
		add		edi,4
		packuswb	mm1,mm1

		dec		ebp
		movd	[esi-4],mm1		;store to both
		movd	[edi-4],mm1
		jne		xloop

		add		esi,eax
		add		edi,ebx
		dec		ecx
		jne		yloop

		pop		ebx
		pop		esi
		pop		edi
		pop		ebp
		emms
		ret
	}
}

void __declspec(naked) dodnr1MMX(unsigned char *dst, unsigned char *src,	int w, int h,
								 int dstmodulo,  __int64 thresh1, __int64 thresh2)
{
	static const __int64 bythree = 0x5555555555555555i64;

	__asm {
		push	ebp
		push	edi
		push	esi
		push	ebx

		mov		edi,[esp+4+16]		;dst
		mov		esi,[esp+8+16]		;src
		mov		edx,[esp+12+16]		;w
		mov		ecx,[esp+16+16]		;h
		xor		eax,eax
		mov		ebx,[esp+20+16]		;dstmodulo
		pxor	mm7,mm7
		movq	mm5,[esp+24+16]		;thresh1
		movq	mm6,[esp+32+16]		;thresh2

yloop:
		mov		ebp,edx
xloop:
		movd	mm0,[esi]		;previous
		pxor	mm7,mm7
		movd	mm1,[edi]		;current
		punpcklbw	mm0,mm7		;mm0 = previous
		punpcklbw	mm1,mm7		;mm1 = current
		movq	mm2,mm0
		movq	mm4,mm1			;mm4 = current
		movq	mm7,mm0
		movq	mm3,mm1
		paddw	mm4,mm4			;mm4 = current * 2
		psubusb	mm2,mm1
		paddw	mm4,mm7			;mm4 = current * 2 + prevvious
		psubusb	mm3,mm7
		pmulhw	mm4,bythree		;mm4 = (current * 2 + previous) / 3
		por		mm2,mm3
		movq	mm3,mm2
		pcmpgtw	mm2,mm5			;set if diff > thresh1
		pcmpgtw	mm3,mm6			;set if diff > thresh2

		;	mm2		mm3		meaning						mm1		mm0		mm4
		;	FALSE	FALSE	diff <= thresh1				off		on		off
		;	FALSE	TRUE	impossible
		;	TRUE	FALSE	thresh1 < diff <= thresh2	off		off		on
		;	TRUE	TRUE	diff > thresh2				on		off		off

		pand	mm1,mm3			;keep pixels exceeding threshold2
		pand	mm4,mm2			;	average pixels <= threshold2...
		pandn	mm2,mm0			;replace pixels below threshold1
		pandn	mm3,mm4			;	but >= threshold1...
		por		mm1,mm2
		add		esi,4
		por		mm1,mm3
		add		edi,4
		packuswb	mm1,mm1

		dec		ebp
		movd	[esi-4],mm1		;store to both
		movd	[edi-4],mm1
		jne		xloop

		add		esi,eax
		add		edi,ebx
		dec		ecx
		jne		yloop

		pop		ebx
		pop		esi
		pop		edi
		pop		ebp
		emms
		ret
	}
}

void DNR::CopyFrame(const unsigned char *dst, const unsigned char *src,
					 int w, int h, int pitch, int rowsize)
{

	int i;
	for (i = 0; i < h; i++)
	{
		memcpy((void *)dst, src, rowsize);
		src += pitch;
		dst += rowsize;
	}
}

PVideoFrame __stdcall DNR::GetFrame(int frame, IScriptEnvironment* env)
{
    unsigned char *curr,*fcrp,*op,a,b,diff;
	int				ia,ib,ic;
	int				x,y,wby4;
	int				srcmodulo;
	int				lowerthreshby2;

	fc = child->GetFrame(frame, env);
	env->MakeWritable(&fc);
	pitch = fc->GetPitch();
	w = fc->GetRowSize();
	h = fc->GetHeight();
	wby4 = w / 4;
	srcmodulo = pitch - w;
	
	fcrp = (unsigned char *) fc->GetReadPtr();
	curr =  (unsigned char *)fcrp;
	op = old_data;

	if(OldFrameNo + 1 != frame) 
	{
		if (!frame)		fb = fc;
		else			fb = child->GetFrame( frame - 1, env);
		CopyFrame( old_data, fb->GetReadPtr(), w, h, pitch, vi.RowSize());
	}
	OldFrameNo = frame;

	if( env->GetCPUFlags() & CPUF_MMX )
	{
		// if YUY2 then comparing each 2 pixels pair
		// if RGB32 then comparing each 1 pixel 
		if(dnr_type)	dodnr1MMX(fcrp, old_data, wby4, h, srcmodulo, thresh1, thresh2);
		else			dodnr2MMX(fcrp, old_data, wby4, h, srcmodulo, thresh1, thresh2);
	}
	else
	{
		if (vi.IsYUY2()) {
			for (y = 0; y < h; y++)	{
				for (x=0; x<wby4; x++) {
					a = *curr;
					b = *op;
					diff = abs (a - b);
					lowerthreshby2 = (dnr_type) ? b : ((a+b+b+b+2)>>2);
					if (diff < threshY_hi) {
						if (diff > threshY_lo)	a = divTable [a + a + b];
						else					a = lowerthreshby2;
					}
					*curr++ = *op++ = a;

					a = *curr;
					b = *op;
					diff = abs (a - b);
					lowerthreshby2 = (dnr_type) ? b : ((a+b+b+b+2)>>2);
					if (diff < threshU_hi) {
						if (diff > threshU_lo)	a = divTable [a + a + b];
						else					a = lowerthreshby2;
					}
					*curr++ = *op++ = a;

					a = *curr;
					b = *op;
					diff = abs (a - b);
					lowerthreshby2 = (dnr_type) ? b : ((a+b+b+b+2)>>2);
					if (diff < threshY_hi) {
						if (diff > threshY_lo)	a = divTable [a + a + b];
						else					a = lowerthreshby2;
					}
					*curr++ = *op++ = a;

					a = *curr;
					b = *op;
					diff = abs (a - b);
					lowerthreshby2 = (dnr_type) ? b : ((a+b+b+b+2)>>2);
					if (diff < threshV_hi) {
						if (diff > threshV_lo)	a = divTable [a + a + b];
						else					a = lowerthreshby2;
					}
					*curr++ = *op++ = a;
				}
				curr += srcmodulo;
			}
		}
		else {	//RGB32
			for (y = 0; y < h; y++) {
				for (x = 0; x < wby4; x++) {
					ia = *(unsigned long *)curr;
					ib = *(unsigned long *)op;
				
					a = (unsigned char)(ia >> 16);
					b = (unsigned char)(ib >> 16);
					diff = abs (a - b);
					lowerthreshby2 = (dnr_type) ? b : ((a+b+b+b+2)>>2);
					if (diff < threshY_hi) {
						if (diff > threshY_lo)	a = divTable [a + a + b];
						else						a = lowerthreshby2;
					}
					ic = (int)a<<8;

					a = (unsigned char)(ia >> 8);
					b = (unsigned char)(ib >> 8);
					diff = abs (a - b);
					lowerthreshby2 = (dnr_type) ? b : ((a+b+b+b+2)>>2);
					if (diff < threshY_hi) {
						if (diff > threshY_lo)	a = divTable [a + a + b];
						else					a = lowerthreshby2;
					}
					ic |= (int)a;
					ic <<= 8;

					a = (unsigned char)ia;
					b = (unsigned char)ib;
					diff = abs (a - b);
					lowerthreshby2 = (dnr_type) ? b : ((a+b+b+b+2)>>2);
					if (diff < threshY_hi) {
						if (diff > threshY_lo)	a = divTable [a + a + b];
						else					a = lowerthreshby2;
					}
					ic |= (int)a;
					*(unsigned long *)curr = *(unsigned long *)op = ic;
					curr += 4; op += 4;
				}
				curr += srcmodulo;
			}
		}
	}
	return fc;
}

bool CheckParm(const char *_type, int *dnr_type)
{
	int i;
	static const char	*TypeTbl[] = { "", "dnr1", "dnr2" };

	*dnr_type = -1;
	if (_type) {
		for (i=0; i<sizeof(TypeTbl)/sizeof(TypeTbl[0]) ; i++) {
			if (!lstrcmpi(_type, TypeTbl[i])) {
				*dnr_type = i;
				break;
			}
		}
	} else {		
		*dnr_type = 0;
	}
	if ( *dnr_type < 0 )	return FALSE;
	if ( *dnr_type == 2)	*dnr_type = 0;
	return TRUE;
}

AVSValue __cdecl Create_DNR(AVSValue args, void* user_data, IScriptEnvironment* env) {
	
	PClip clip = args[0].AsClip();
	const VideoInfo& vi = clip->GetVideoInfo();
	int dnr_type;

	if ( !CheckParm(args[7].AsString(0), &dnr_type) )
		env->ThrowError("DNR2: parameter error : type");

	if(vi.IsRGB32() || vi.IsYUY2())
		return new DNR(clip,
						args[1].AsInt(8),		// threshY_hi
						args[2].AsInt(-1),		// threshY_lo
						args[3].AsInt(-1),		// threshU_hi
						args[4].AsInt(-1),		// threshU_lo
						args[5].AsInt(-1),		// threshV_hi
						args[6].AsInt(-1),		// threshV_lo
						dnr_type,				// type
						env);
#ifdef AVISYNTH_2_5
	if(vi.IsYV12()) {
		return new DNR_YV12(clip,
							args[1].AsInt(8),		// threshY_hi
							args[2].AsInt(-1),		// threshY_lo
							args[3].AsInt(-1),		// threshU_hi
							args[4].AsInt(-1),		// threshU_lo
							args[5].AsInt(-1),		// threshV_hi
							args[6].AsInt(-1),		// threshV_lo
							dnr_type,				// type
							env);
	} else {
		env->ThrowError("DNR2: requires YUY2 or RGB32 or YV12(or I420) input");
	}
#endif
	env->ThrowError("DNR2: requires YUY2 or RGB32 input");
	return 0;
}

extern "C" __declspec(dllexport) const char* __stdcall AVISYNTHPLUGININIT(IScriptEnvironment* env)
{
    env->AddFunction("DNR2",
					 "c[threshY_hi]i[threshY_lo]i[threshU_hi]i[threshU_lo]i[threshV_hi]i[threshV_lo]i[type]s",
		             Create_DNR, 0);
	return 0;
}
