//  DNR2 Ver.0.7 : Dynamic Noise Reduction 2 for Avisynth
//  Copyright (C) 2002 kiraru2002
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA,
//
//  or visit http://www.gnu.org/copyleft/gpl.html .
//
//------------------------------------------------------------------------------------
// kiraru2002's web site  http://members.tripod.co.jp/kiraru2002/
//
// This plug-in is ported from the VirtualDub's plug-ins "DNR1" and
// "DNR2" which Steven Don and Avery Lee created.
// You MUST NOT contact to Steven Don and Avery Lee for any trouble
// which happened using this plug-in, and for any things about this plug-in.
//------------------------------------------------------------------------------------
#include "Dnr.h"

#ifdef AVISYNTH_2_5

DNR_YV12::DNR_YV12(PClip _child,	int _threshY_hi, int _threshY_lo,
							int _threshU_hi, int _threshU_lo,
							int _threshV_hi, int _threshV_lo,
							int _dnr_type,
						IScriptEnvironment* env) :
					GenericVideoFilter(_child),
					threshY_hi(_threshY_hi),threshY_lo(_threshY_lo),
					threshU_hi(_threshU_hi),threshU_lo(_threshU_lo),
					threshV_hi(_threshV_hi),threshV_lo(_threshV_lo),dnr_type(_dnr_type)
{

	if(vi.width & 7)	env->ThrowError("DNR: width must be multiple of 8");

	if (threshU_hi == -1)	threshU_hi = threshY_hi * 3 / 4;
	if (threshV_hi == -1)	threshV_hi = threshU_hi;

	if (threshY_lo == -1)	threshY_lo = (threshY_hi>>1) + 1;
	if (threshU_lo == -1)	threshU_lo = (threshU_hi>>1) + 1;
	if (threshV_lo == -1)	threshV_lo = (threshV_hi>>1) + 1;

	thresh1 = 0x0001000100010001i64 * threshY_lo;
	thresh2 = 0x0001000100010001i64 * threshY_hi;
	threshU1 = 0x0001000100010001i64 * threshU_lo;
	threshU2 = 0x0001000100010001i64 * threshU_hi;
	threshV1 = 0x0001000100010001i64 * threshV_lo;
	threshV2 = 0x0001000100010001i64 * threshV_hi;

	OldFrameNo = -9999;

	divTable = NULL;
	if( !(env->GetCPUFlags() & CPUF_MMX) ) {
		if (!(divTable = new unsigned char [766])) return;
		for (unsigned int i = 0; i < 766; i++) 		divTable[i] = i / 3;
	}
		
	fc = child->GetFrame(0, env);
	if (!(old_data[0] = new unsigned char [fc->GetRowSize(PLANAR_Y) * fc->GetHeight(PLANAR_Y)])) return;
	if (!(old_data[1] = new unsigned char [fc->GetRowSize(PLANAR_U) * fc->GetHeight(PLANAR_U)])) return;
	if (!(old_data[2] = new unsigned char [fc->GetRowSize(PLANAR_V) * fc->GetHeight(PLANAR_V)])) return;

}
					

DNR_YV12::~DNR_YV12()
{
	for(int i=0; i<3; i++) {
		if(old_data[i])	{ delete [] old_data[i]; old_data[i] = NULL; }
	}
	if(divTable)	{ delete [] divTable; divTable = NULL; }
}


void DNR_YV12::CopyFrame(PVideoFrame &src, IScriptEnvironment* env)
{
	env->BitBlt(old_data[0], src->GetRowSize(PLANAR_Y),
				src->GetReadPtr(PLANAR_Y), src->GetPitch(PLANAR_Y),
				src->GetRowSize(PLANAR_Y), src->GetHeight(PLANAR_Y) );
	env->BitBlt(old_data[1], src->GetRowSize(PLANAR_U),
				src->GetReadPtr(PLANAR_U), src->GetPitch(PLANAR_U),
				src->GetRowSize(PLANAR_U), src->GetHeight(PLANAR_U) );
	env->BitBlt(old_data[2], src->GetRowSize(PLANAR_V),
				src->GetReadPtr(PLANAR_V), src->GetPitch(PLANAR_V),
				src->GetRowSize(PLANAR_V), src->GetHeight(PLANAR_V) );
}

const fpDODNRMMX dodnrMMX[] = { dodnr2MMX, dodnr1MMX };

PVideoFrame __stdcall DNR_YV12::GetFrame(int frame, IScriptEnvironment* env)
{
    unsigned char *srcp, *oldp, a,b,diff;
	int				i,x,y;
	int				srcmodulo;
	int				lowerthreshby2;
	int				planar, Planar_Tbl[3] = { PLANAR_Y, PLANAR_U, PLANAR_V };

	fc = child->GetFrame(frame, env);
	env->MakeWritable(&fc);

	if(OldFrameNo + 1 != frame) 
	{
		if (!frame)		fb = fc;
		else			fb = child->GetFrame( frame-1 , env);
		CopyFrame(fb, env);
	}
	OldFrameNo = frame;

	if( env->GetCPUFlags() & CPUF_MMX )
	{
//		if(fc->GetReadPtr(PLANAR_Y) != fc->GetWritePtr(PLANAR_Y)) {
//			char buf[256];
//			sprintf(buf,"DNR2: readptr=%x writeptr=%x unmatched!", fc->GetReadPtr(PLANAR_Y),fc->GetWritePtr(PLANAR_Y));
//			env->ThrowError(buf);
//		}

		dodnrMMX[dnr_type]((unsigned char *)fc->GetReadPtr(PLANAR_Y), old_data[0],
							fc->GetRowSize(PLANAR_Y)/4,	fc->GetHeight(PLANAR_Y),
							fc->GetPitch(PLANAR_Y) - fc->GetRowSize(PLANAR_Y),
							thresh1, thresh2);
		dodnrMMX[dnr_type]((unsigned char *)fc->GetReadPtr(PLANAR_U), old_data[1],
							fc->GetRowSize(PLANAR_U)/4,	fc->GetHeight(PLANAR_U),
							fc->GetPitch(PLANAR_U) - fc->GetRowSize(PLANAR_U),
							threshU1, threshU2);
		dodnrMMX[dnr_type]((unsigned char *)fc->GetReadPtr(PLANAR_V), old_data[2],
							fc->GetRowSize(PLANAR_V)/4,	fc->GetHeight(PLANAR_V),
							fc->GetPitch(PLANAR_V) - fc->GetRowSize(PLANAR_V),
							threshV1, threshV2);
	}
	else
	{
		for(i=0; i < 3; i++) {
			planar = Planar_Tbl[i];
			pitch = fc->GetPitch(planar);
			w = fc->GetRowSize(planar);
			h = fc->GetHeight(planar);
			srcmodulo = pitch - w;
	
			srcp = (unsigned char *) fc->GetReadPtr(planar);
			oldp = old_data[i];

			for (y = 0; y < h; y++)	{
				for (x=0; x < w; x++) {
					a = *srcp;
					b = *oldp;
					diff = abs (a - b);
					lowerthreshby2 = (dnr_type) ? b : ((a+b+b+b+2)>>2);
					if (diff < threshY_hi) {
						if (diff > threshY_lo)	a = divTable [a + a + b];
						else					a = lowerthreshby2;
					}
					*srcp++ = *oldp++ = a;
				}
				srcp += srcmodulo;
			}
		}
	}
	return fc;
}

#endif	// AVISYNTH_2_5
